import asyncio
import json
import logging
import multiprocessing
import os
import signal
import socket
import sys
import time
from collections.abc import Generator
from pathlib import Path
from typing import AsyncGenerator, Dict, List, Optional, Union

import common.errors as errors
import shared.util as utils
import uvicorn
from common.termination_handler_middleware import TerminationHandlerMiddleware
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import ORJSONResponse, StreamingResponse
from fastapi.routing import APIRoute as FastAPIRoute
from model_wrapper import ModelWrapper
from shared.logging import setup_logging
from shared.serialization import (
    DeepNumpyEncoder,
    truss_msgpack_deserialize,
    truss_msgpack_serialize,
)
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import ClientDisconnect
from starlette.responses import Response

# [IMPORTANT] A lot of things depend on this currently.
# Please consider the following when increasing this:
# 1. Self-termination on model load fail.
# 2. Graceful termination.
DEFAULT_NUM_WORKERS = 1
DEFAULT_NUM_SERVER_PROCESSES = 1
WORKER_TERMINATION_TIMEOUT_SECS = 120.0
WORKER_TERMINATION_CHECK_INTERVAL_SECS = 0.5


async def parse_body(request: Request) -> bytes:
    """
    Used by FastAPI to read body in an asynchronous manner
    """
    try:
        return await request.body()
    except ClientDisconnect as exc:
        error_message = "Client disconnected"
        logging.error(error_message)
        raise HTTPException(status_code=499, detail=error_message) from exc


FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(levelname)s [%(funcName)s():%(lineno)s] %(message)s"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
INFERENCE_SERVER_FAILED_FILE = Path("~/inference_server_crashed.txt").expanduser()
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt=DATE_FORMAT)


class UvicornCustomServer(multiprocessing.Process):
    def __init__(
        self, config: uvicorn.Config, sockets: Optional[List[socket.socket]] = None
    ):
        super().__init__()
        self.sockets = sockets
        self.config = config

    def stop(self):
        self.terminate()

    def run(self):
        server = uvicorn.Server(config=self.config)
        asyncio.run(server.serve(sockets=self.sockets))


class BasetenEndpoints:
    """The implementation of the model server endpoints.

    Historically, we relied on the kserve server interface, which assumes that
    multiple models are running behind a registry. As a result, some arguments to
    to functions will rename unused except for backwards compatibility checks.
    """

    def __init__(self, model: ModelWrapper) -> None:
        self._model = model

    def _safe_lookup_model(self, model_name: str) -> ModelWrapper:
        if model_name != self._model.name:
            raise errors.ModelMissingError(model_name)
        return self._model

    @staticmethod
    def check_healthy(model: ModelWrapper):
        if model.load_failed():
            INFERENCE_SERVER_FAILED_FILE.touch()
            os.kill(os.getpid(), signal.SIGKILL)

        if not model.ready:
            raise errors.ModelNotReady(model.name)

    async def model_ready(self, model_name: str) -> Dict[str, Union[str, bool]]:
        self.check_healthy(self._safe_lookup_model(model_name))

        return {}

    async def invocations_ready(self) -> Dict[str, Union[str, bool]]:
        """
        This method provides compatibility with Sagemaker hosting for the 'ping' endpoint.
        """
        if self._model is None:
            raise errors.ModelMissingError("model")
        self.check_healthy(self._model)

        return {}

    async def invocations(
        self, request: Request, body_raw: bytes = Depends(parse_body)
    ) -> Response:
        """
        This method provides compatibility with Sagemaker hosting for the 'invocations' endpoint.
        """
        return await self.predict(self._model.name, request, body_raw)

    async def predict(
        self, model_name: str, request: Request, body_raw: bytes = Depends(parse_body)
    ) -> Response:
        """
        This method calls the user-provided predict method
        """
        model: ModelWrapper = self._safe_lookup_model(model_name)

        self.check_healthy(model)

        body: Dict
        if self.is_binary(request):
            body = truss_msgpack_deserialize(body_raw)
        else:
            try:
                body = json.loads(body_raw)
            except json.JSONDecodeError as e:
                error_message = f"Invalid JSON payload: {str(e)}"
                logging.error(error_message)
                raise HTTPException(status_code=400, detail=error_message)

        # calls ModelWrapper.__call__, which runs validate, preprocess, predict, and postprocess
        response: Union[Dict, Generator] = await model(
            body,
            headers=utils.transform_keys(request.headers, lambda key: key.lower()),
        )

        # In the case that the model returns a Generator object, return a
        # StreamingResponse instead.
        if isinstance(response, (AsyncGenerator, Generator)):
            # media_type in StreamingResponse sets the Content-Type header
            return StreamingResponse(response, media_type="application/octet-stream")

        response_headers = {}
        if self.is_binary(request):
            response_headers["Content-Type"] = "application/octet-stream"
            return Response(
                content=truss_msgpack_serialize(response), headers=response_headers
            )
        else:
            response_headers["Content-Type"] = "application/json"
            return Response(
                content=json.dumps(response, cls=DeepNumpyEncoder),
                headers=response_headers,
            )

    async def schema(self, model_name: str) -> Dict:
        model: ModelWrapper = self._safe_lookup_model(model_name)

        if model.truss_schema is None:
            print("DID NOT FIND MODEL SCHEMA")
            # If there is not a TrussSchema, we return a 404.

            if model.ready:
                raise HTTPException(status_code=404, detail="No schema found")
            else:
                raise HTTPException(
                    status_code=503,
                    detail="Schema not available, please try again later.",
                )
        else:
            return model.truss_schema.serialize()

    @staticmethod
    def is_binary(request: Request):
        return (
            "Content-Type" in request.headers
            and request.headers["Content-Type"] == "application/octet-stream"
        )


class TrussServer:
    """This wrapper class manages creation and cleanup of uvicorn server processes running the FastAPI inference server app

    TrussServer runs as a main process managing UvicornCustomServer subprocesses that in turn may manage
    their own worker processes. Notably, this main process is kept alive when running `servers_task()`
    because of the child uvicorn server processes' main loop.

    """

    def __init__(
        self,
        http_port: int,
        config: Dict,
        setup_json_logger: bool = True,
    ):
        self.http_port = http_port
        self._config = config
        self._model = ModelWrapper(self._config)
        self._endpoints = BasetenEndpoints(self._model)
        self._setup_json_logger = setup_json_logger

    def cleanup(self):
        if INFERENCE_SERVER_FAILED_FILE.exists():
            INFERENCE_SERVER_FAILED_FILE.unlink()

    def on_startup(self):
        """
        This method will be started inside the main process, so here is where we want to setup our logging and model
        """
        self.cleanup()

        if self._setup_json_logger:
            setup_logging()

        self._model.start_load()

    def create_application(self):
        app = FastAPI(
            title="Baseten Inference Server",
            docs_url=None,
            redoc_url=None,
            default_response_class=ORJSONResponse,
            on_startup=[self.on_startup],
            routes=[
                # liveness endpoint
                FastAPIRoute(r"/", lambda: True),
                # readiness endpoint
                FastAPIRoute(
                    r"/v1/models/{model_name}", self._endpoints.model_ready, tags=["V1"]
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}/schema",
                    self._endpoints.schema,
                    methods=["GET"],
                    tags=["V1"],
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:predict",
                    self._endpoints.predict,
                    methods=["POST"],
                    tags=["V1"],
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:predict_binary",
                    self._endpoints.predict,
                    methods=["POST"],
                    tags=["V1"],
                ),
                # Endpoint aliases for Sagemaker hosting
                FastAPIRoute(r"/ping", self._endpoints.invocations_ready),
                FastAPIRoute(
                    r"/invocations",
                    self._endpoints.invocations,
                    methods=["POST"],
                ),
            ],
            exception_handlers={
                errors.InferenceError: errors.inference_error_handler,
                errors.ModelNotFound: errors.model_not_found_handler,
                errors.ModelNotReady: errors.model_not_ready_handler,
                NotImplementedError: errors.not_implemented_error_handler,
                HTTPException: errors.http_exception_handler,
                Exception: errors.generic_exception_handler,
            },
        )

        def exit_self():
            # Note that this kills the current process, the worker process, not
            # the main truss_server process.
            utils.kill_child_processes(os.getpid())
            sys.exit()

        termination_handler_middleware = TerminationHandlerMiddleware(
            on_stop=lambda: None,
            on_term=exit_self,
        )
        app.add_middleware(BaseHTTPMiddleware, dispatch=termination_handler_middleware)
        return app

    def start(self):
        cfg = uvicorn.Config(
            self.create_application(),
            # We hard-code the http parser as h11 (the default) in case the user has
            # httptools installed, which does not work with our requests & version
            # of uvicorn.
            http="h11",
            host="0.0.0.0",
            port=self.http_port,
            workers=DEFAULT_NUM_WORKERS,
            log_config={
                "version": 1,
                "formatters": {
                    "default": {
                        "()": "uvicorn.logging.DefaultFormatter",
                        "datefmt": DATE_FORMAT,
                        "fmt": "%(asctime)s.%(msecs)03d %(name)s %(levelprefix)s %(message)s",
                        "use_colors": None,
                    },
                    "access": {
                        "()": "uvicorn.logging.AccessFormatter",
                        "datefmt": DATE_FORMAT,
                        "fmt": "%(asctime)s.%(msecs)03d %(name)s %(levelprefix)s %(client_addr)s %(process)s - "
                        '"%(request_line)s" %(status_code)s',
                        # noqa: E501
                    },
                },
                "handlers": {
                    "default": {
                        "formatter": "default",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stderr",
                    },
                    "access": {
                        "formatter": "access",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",
                    },
                },
                "loggers": {
                    "uvicorn": {"handlers": ["default"], "level": "INFO"},
                    "uvicorn.error": {"level": "INFO"},
                    "uvicorn.access": {
                        "handlers": ["access"],
                        "level": "INFO",
                        "propagate": False,
                    },
                },
            },
        )

        # Call this so uvloop gets used
        cfg.setup_event_loop()

        async def serve():
            serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            serversocket.bind((cfg.host, cfg.port))
            serversocket.listen(5)

            num_server_procs = self._config.get("runtime", {}).get(
                "num_workers", DEFAULT_NUM_SERVER_PROCESSES
            )
            logging.info(f"starting {num_server_procs} uvicorn server processes")
            servers: List[UvicornCustomServer] = []
            for _ in range(num_server_procs):
                server = UvicornCustomServer(config=cfg, sockets=[serversocket])
                server.start()
                servers.append(server)

            def stop_servers():
                # Send stop signal, then wait for all to exit
                for server in servers:
                    # Sends term signal to the process, which should be handled
                    # by the termination handler.
                    server.stop()

                termination_check_attempts = int(
                    WORKER_TERMINATION_TIMEOUT_SECS
                    / WORKER_TERMINATION_CHECK_INTERVAL_SECS
                )
                for _ in range(termination_check_attempts):
                    time.sleep(WORKER_TERMINATION_CHECK_INTERVAL_SECS)
                    if utils.all_processes_dead(servers):
                        return

            for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGQUIT]:
                signal.signal(sig, lambda sig, frame: stop_servers())

        async def servers_task():
            servers = [serve()]
            await asyncio.gather(*servers)

        asyncio.run(servers_task())



import json
import os
import subprocess
import time
from pathlib import Path
from typing import AsyncGenerator, Optional

import tritonclient.grpc.aio as grpcclient
import tritonclient.http as httpclient
from constants import (
    ENTRYPOINT_MODEL_NAME,
    GRPC_SERVICE_PORT,
    TENSORRT_LLM_MODEL_REPOSITORY_PATH,
)
from schema import ModelInput
from utils import download_engine, prepare_model_repository


class TritonServer:
    def __init__(self, grpc_port: int = 8001, http_port: int = 8003):
        self.grpc_port = grpc_port
        self.http_port = http_port
        self._server_process = None

    def create_model_repository(
        self,
        truss_data_dir: Path,
        engine_repository_path: Optional[str] = None,
        huggingface_auth_token: Optional[str] = None,
    ) -> None:
        if engine_repository_path:
            download_engine(
                engine_repository=engine_repository_path,
                fp=truss_data_dir,
                auth_token=huggingface_auth_token,
            )
        prepare_model_repository(truss_data_dir)
        return

    def start(self, tensor_parallelism: int = 1, env: dict = {}) -> None:
        def build_server_start_command() -> list:
            """
            Triton Inference Server has different startup commands depending on the tensor
            parallelism (TP) configuration. This function starts the server with the appropriate
            command.
            """
            base_command = [
                "tritonserver",
                f"--model-repository={TENSORRT_LLM_MODEL_REPOSITORY_PATH}",
                f"--grpc-port={str(self.grpc_port)}",
                f"--http-port={str(self.http_port)}",
                # "--log-verbose=1",
            ]

            if tensor_parallelism == 1:
                return base_command

            mpirun_command = ["mpirun", "--allow-run-as-root"]
            mpi_commands = []
            for i in range(tensor_parallelism):
                mpi_command = [
                    "-n=1",
                    *base_command,
                    "--disable-auto-complete-config",
                    f"--backend-config=python,shm-region-prefix-name=prefix{str(i)}_",
                ]
                mpi_commands.append(" ".join(mpi_command))

            combined_mpi_commands = " : ".join(mpi_commands)
            return mpirun_command + [combined_mpi_commands]

        command = build_server_start_command()
        self._server_process = subprocess.Popen(  # type: ignore
            command,
            env={**os.environ, **env},
        )
        while not self.is_alive and not self.is_ready:
            time.sleep(2)
        return

    def stop(self):
        if self._server_process:
            if self.is_server_ready:
                self._server_process.kill()
            self._server_process = None
        return

    @property
    def is_alive(self) -> bool:
        try:
            http_client = httpclient.InferenceServerClient(
                url=f"localhost:{self.http_port}", verbose=False
            )
            return http_client.is_server_live()
        except ConnectionRefusedError:
            return False

    @property
    def is_ready(self) -> bool:
        try:
            http_client = httpclient.InferenceServerClient(
                url=f"localhost:{self.http_port}", verbose=False
            )
            return http_client.is_model_ready(model_name=ENTRYPOINT_MODEL_NAME)
        except ConnectionRefusedError:
            return False


class TritonClient:
    def __init__(self, grpc_service_port: int = GRPC_SERVICE_PORT):
        self.grpc_service_port = grpc_service_port
        self._grpc_client = None

    def start_grpc_stream(self) -> grpcclient.InferenceServerClient:
        if self._grpc_client:
            return self._grpc_client

        self._grpc_client = grpcclient.InferenceServerClient(
            url=f"localhost:{self.grpc_service_port}", verbose=False
        )
        return self._grpc_client

    async def infer(
        self, model_input: ModelInput, model_name="ensemble"
    ) -> AsyncGenerator[str, None]:
        grpc_client_instance = self.start_grpc_stream()
        inputs = model_input.to_tensors()

        async def input_generator():
            yield {
                "model_name": model_name,
                "inputs": inputs,
                "request_id": model_input.request_id,
            }

        response_iterator = grpc_client_instance.stream_infer(
            inputs_iterator=input_generator(),
        )

        try:
            async for response in response_iterator:
                result, error = response
                if result:
                    result = result.as_numpy("text_output")
                    yield result[0].decode("utf-8")
                else:
                    yield json.dumps({"status": "error", "message": error.message()})

        except grpcclient.InferenceServerException as e:
            print(f"InferenceServerException: {e}")



from typing import Callable, Dict, Optional, TypeVar

X = TypeVar("X")
Y = TypeVar("Y")
Z = TypeVar("Z")


def transform_optional(x: Optional[X], fn: Callable[[X], Optional[Y]]) -> Optional[Y]:
    if x is None:
        return None

    return fn(x)


def transform_keys(d: Dict[X, Z], fn: Callable[[X], Y]) -> Dict[Y, Z]:
    return {fn(key): value for key, value in d.items()}



import logging
import re
from pathlib import Path
from typing import List, Optional, Tuple

import click
import rich
import yaml
from requests import ReadTimeout
from truss.local.local_config_handler import LocalConfigHandler
from truss.remote.baseten.api import BasetenApi
from truss.remote.baseten.auth import AuthService
from truss.remote.baseten.core import (
    ModelId,
    ModelIdentifier,
    ModelName,
    ModelVersionId,
    archive_truss,
    create_truss_service,
    exists_model,
    get_dev_version,
    get_dev_version_from_versions,
    get_model_versions,
    get_prod_version_from_versions,
    upload_truss,
)
from truss.remote.baseten.error import ApiError
from truss.remote.baseten.service import BasetenService
from truss.remote.baseten.utils.transfer import base64_encoded_json_str
from truss.remote.truss_remote import TrussRemote, TrussService
from truss.truss_config import ModelServer
from truss.truss_handle import TrussHandle
from truss.util.path import is_ignored, load_trussignore_patterns
from watchfiles import watch


class BasetenRemote(TrussRemote):
    def __init__(self, remote_url: str, api_key: str, **kwargs):
        super().__init__(remote_url, **kwargs)
        self._auth_service = AuthService(api_key=api_key)
        self._api = BasetenApi(f"{self._remote_url}/graphql/", self._auth_service)

    def authenticate(self):
        return self._auth_service.validate()

    def push(  # type: ignore
        self,
        truss_handle: TrussHandle,
        model_name: str,
        publish: bool = True,
        trusted: bool = False,
        promote: bool = False,
        preserve_previous_prod_deployment: bool = False,
        deployment_name: Optional[str] = None,
    ):
        if model_name.isspace():
            raise ValueError("Model name cannot be empty")

        model_id = exists_model(self._api, model_name)

        gathered_truss = TrussHandle(truss_handle.gather())
        if gathered_truss.spec.model_server != ModelServer.TrussServer:
            publish = True

        if promote:
            # If we are promoting a model after deploy, it must be published.
            # Draft models cannot be promoted.
            publish = True

        if not publish and deployment_name:
            raise ValueError(
                "Deployment name cannot be used for development deployment"
            )

        if not promote and preserve_previous_prod_deployment:
            raise ValueError(
                "preserve-previous-production-deployment can only be used with the '--promote' option"
            )

        if deployment_name and not re.match(r"^[0-9a-zA-Z_\-\.]*$", deployment_name):
            raise ValueError(
                "Deployment name must only contain alphanumeric, -, _ and . characters"
            )

        encoded_config_str = base64_encoded_json_str(
            gathered_truss._spec._config.to_dict()
        )

        temp_file = archive_truss(gathered_truss)
        s3_key = upload_truss(self._api, temp_file)

        model_id, model_version_id = create_truss_service(
            api=self._api,
            model_name=model_name,
            s3_key=s3_key,
            config=encoded_config_str,
            is_draft=not publish,
            model_id=model_id,
            is_trusted=trusted,
            promote=promote,
            preserve_previous_prod_deployment=preserve_previous_prod_deployment,
            deployment_name=deployment_name,
        )

        return BasetenService(
            model_id=model_id,
            model_version_id=model_version_id,
            is_draft=not publish,
            api_key=self._auth_service.authenticate().value,
            service_url=f"{self._remote_url}/model_versions/{model_version_id}",
            truss_handle=truss_handle,
        )

    @staticmethod
    def _get_matching_version(model_versions: List[dict], published: bool) -> dict:
        if not published:
            # Return the development model version.
            dev_version = get_dev_version_from_versions(model_versions)
            if not dev_version:
                raise click.UsageError(
                    "No development model found. Run `truss push` then try again."
                )
            return dev_version

        # Return the production deployment version.
        prod_version = get_prod_version_from_versions(model_versions)
        if not prod_version:
            raise click.UsageError(
                "No production model found. Run `truss push --publish` then try again."
            )
        return prod_version

    @staticmethod
    def _get_service_url_path_and_model_ids(
        api: BasetenApi, model_identifier: ModelIdentifier, published: bool
    ) -> Tuple[str, str, str]:
        if isinstance(model_identifier, ModelVersionId):
            try:
                model_version = api.get_model_version_by_id(model_identifier.value)
            except ApiError:
                raise click.UsageError(
                    f"Model version {model_identifier.value} not found."
                )
            model_version_id = model_version["model_version"]["id"]
            model_id = model_version["model_version"]["oracle"]["id"]
            service_url_path = f"/model_versions/{model_version_id}"
            return service_url_path, model_id, model_version_id

        if isinstance(model_identifier, ModelName):
            model_id, model_versions = get_model_versions(api, model_identifier)
            model_version = BasetenRemote._get_matching_version(
                model_versions, published
            )
            model_version_id = model_version["id"]
            service_url_path = f"/model_versions/{model_version_id}"
        elif isinstance(model_identifier, ModelId):
            # TODO(helen): consider making this consistent with getting the
            # service via model_name / respect --published in service_url_path.
            try:
                model = api.get_model_by_id(model_identifier.value)
            except ApiError:
                raise click.UsageError(f"Model {model_identifier.value} not found.")
            model_id = model["model"]["id"]
            model_version_id = model["model"]["primary_version"]["id"]
            service_url_path = f"/models/{model_id}"
        else:
            # Model identifier is of invalid type.
            raise click.UsageError(
                "You must either be inside of a Truss directory, or provide --model-deployment or --model options."
            )

        return service_url_path, model_id, model_version_id

    def get_service(self, **kwargs) -> BasetenService:
        try:
            model_identifier = kwargs["model_identifier"]
        except KeyError:
            raise ValueError("Baseten Service requires a model_identifier")

        published = kwargs.get("published", False)
        (
            service_url_path,
            model_id,
            model_version_id,
        ) = self._get_service_url_path_and_model_ids(
            self._api, model_identifier, published
        )

        return BasetenService(
            model_id=model_id,
            model_version_id=model_version_id,
            is_draft=not published,
            api_key=self._auth_service.authenticate().value,
            service_url=f"{self._remote_url}{service_url_path}",
        )

    def get_remote_logs_url(
        self,
        service: TrussService,
    ) -> str:
        return service.logs_url(self._remote_url)

    def sync_truss_to_dev_version_by_name(
        self,
        model_name: str,
        target_directory: str,
    ):
        # verify that development deployment exists for given model name
        dev_version = get_dev_version(
            self._api, model_name
        )  # pylint: disable=protected-access
        if not dev_version:
            raise click.UsageError(
                "No development model found. Run `truss push` then try again."
            )

        watch_path = Path(target_directory)
        trussignore_patterns = load_trussignore_patterns()

        def watch_filter(_, path):
            return not is_ignored(
                Path(path),
                trussignore_patterns,
            )

        # disable watchfiles logger
        logging.getLogger("watchfiles.main").disabled = True

        rich.print(f"🚰 Attempting to sync truss at '{watch_path}' with remote")
        self.patch(watch_path)

        rich.print(f"👀 Watching for changes to truss at '{watch_path}' ...")
        for _ in watch(watch_path, watch_filter=watch_filter, raise_interrupt=False):
            self.patch(watch_path)

    def patch(
        self,
        watch_path: Path,
    ):
        from truss.cli.console import console, error_console

        try:
            truss_handle = TrussHandle(watch_path)
        except yaml.parser.ParserError:
            error_console.print("Unable to parse config file")
            return
        except ValueError:
            error_console.print(f"Error when reading truss from directory {watch_path}")
            return
        model_name = truss_handle.spec.config.model_name
        dev_version = get_dev_version(self._api, model_name)  # type: ignore
        if not dev_version:
            error_console.print(
                f"No development deployment found with model name: {model_name}"
            )
            return
        truss_hash = dev_version.get("truss_hash", None)
        truss_signature = dev_version.get("truss_signature", None)
        if not (truss_hash and truss_signature):
            error_console.print(
                """Failed to inspect a running remote deployment to watch for changes.
Ensure that there exists a running remote deployment before attempting to watch for changes
            """
            )
            return
        LocalConfigHandler.add_signature(truss_hash, truss_signature)
        try:
            patch_request = truss_handle.calc_patch(truss_hash)
        except Exception:
            error_console.print("Failed to calculate patch, bailing on patching")
            return
        if patch_request:
            if (
                patch_request.prev_hash == patch_request.next_hash
                or len(patch_request.patch_ops) == 0
            ):
                console.print("No changes observed, skipping patching")
                return
            try:
                with console.status("Applying patch..."):
                    resp = self._api.patch_draft_truss(model_name, patch_request)
            except ReadTimeout:
                error_console.print(
                    "Read Timeout when attempting to connect to remote. Bailing on patching"
                )
                return
            except Exception:
                error_console.print(
                    "Failed to patch draft deployment, bailing on patching"
                )
                return
            if not resp["succeeded"]:
                needs_full_deploy = resp.get("needs_full_deploy", None)
                if needs_full_deploy:
                    error_console.print(
                        f"Model {model_name} is not able to be patched, use `truss push` to deploy"
                    )
                else:
                    error_console.print(
                        f"Failed to patch: `{resp['error']}`. Model left in original state"
                    )
            else:
                console.print(
                    resp.get(
                        "success_message",
                        f"Model {model_name} patched successfully",
                    ),
                    style="green",
                )



# Copyright 2023, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import csv
import json
import os
from typing import List

import numpy as np
import triton_python_backend_utils as pb_utils
from transformers import AutoTokenizer, LlamaTokenizer, T5Tokenizer


class TritonPythonModel:
    """Your Python model must use the same class name. Every Python model
    that is created must have "TritonPythonModel" as the class name.
    """

    def initialize(self, args):
        """`initialize` is called only once when the model is being loaded.
        Implementing `initialize` function is optional. This function allows
        the model to initialize any state associated with this model.
        Parameters
        ----------
        args : dict
          Both keys and values are strings. The dictionary keys and values are:
          * model_config: A JSON string containing the model configuration
          * model_instance_kind: A string containing model instance kind
          * model_instance_device_id: A string containing model instance device ID
          * model_repository: Model repository path
          * model_version: Model version
          * model_name: Model name
        """
        # Parse model configs
        model_config = json.loads(args["model_config"])
        # NOTE: Keep this in sync with the truss model.py variable
        tokenizer_dir = os.environ["TRITON_TOKENIZER_REPOSITORY"]
        tokenizer_type = model_config["parameters"]["tokenizer_type"]["string_value"]
        self.add_special_tokens = model_config["parameters"].get(
            "add_special_tokens", {"string_value": "false"}
        )["string_value"].lower() in ["true", "1", "t", "y", "yes"]

        if tokenizer_type == "t5":
            self.tokenizer = T5Tokenizer(vocab_file=tokenizer_dir, padding_side="left")
        elif tokenizer_type == "auto":
            self.tokenizer = AutoTokenizer.from_pretrained(
                tokenizer_dir, padding_side="left"
            )
        elif tokenizer_type == "llama":
            self.tokenizer = LlamaTokenizer.from_pretrained(
                tokenizer_dir, legacy=False, padding_side="left"
            )
        else:
            raise AttributeError(f"Unexpected tokenizer type: {tokenizer_type}")
        self.tokenizer.pad_token = self.tokenizer.eos_token

        self.pad_id = self.tokenizer.encode(
            self.tokenizer.pad_token, add_special_tokens=False
        )[0]

        # Parse model output configs and convert Triton types to numpy types
        input_names = [
            "INPUT_ID",
            "REQUEST_INPUT_LEN",
            "BAD_WORDS_IDS",
            "STOP_WORDS_IDS",
        ]
        for input_name in input_names:
            setattr(
                self,
                input_name.lower() + "_dtype",
                pb_utils.triton_string_to_numpy(
                    pb_utils.get_output_config_by_name(model_config, input_name)[
                        "data_type"
                    ]
                ),
            )

    def execute(self, requests):
        """`execute` must be implemented in every Python model. `execute`
        function receives a list of pb_utils.InferenceRequest as the only
        argument. This function is called when an inference is requested
        for this model. Depending on the batching configuration (e.g. Dynamic
        Batching) used, `requests` may contain multiple requests. Every
        Python model, must create one pb_utils.InferenceResponse for every
        pb_utils.InferenceRequest in `requests`. If there is an error, you can
        set the error argument when creating a pb_utils.InferenceResponse.
        Parameters
        ----------
        requests : list
          A list of pb_utils.InferenceRequest
        Returns
        -------
        list
          A list of pb_utils.InferenceResponse. The length of this list must
          be the same as `requests`
        """

        responses = []

        # Every Python backend must iterate over everyone of the requests
        # and create a pb_utils.InferenceResponse for each of them.
        for idx, request in enumerate(requests):
            # Get input tensors
            query = pb_utils.get_input_tensor_by_name(request, "QUERY").as_numpy()
            request_output_len = pb_utils.get_input_tensor_by_name(
                request, "REQUEST_OUTPUT_LEN"
            ).as_numpy()

            bad_words_dict = pb_utils.get_input_tensor_by_name(
                request, "BAD_WORDS_DICT"
            ).as_numpy()
            stop_words_dict = pb_utils.get_input_tensor_by_name(
                request, "STOP_WORDS_DICT"
            ).as_numpy()

            # Preprocessing input data.
            input_id, request_input_len = self._create_request(query)
            bad_words = self._to_word_list_format(bad_words_dict)
            stop_words = self._to_word_list_format(stop_words_dict)

            # Create output tensors. You need pb_utils.Tensor
            # objects to create pb_utils.InferenceResponse.
            input_id_tensor = pb_utils.Tensor(
                "INPUT_ID", np.array(input_id).astype(self.input_id_dtype)
            )
            request_input_len_tensor = pb_utils.Tensor(
                "REQUEST_INPUT_LEN",
                np.array(request_input_len).astype(self.request_input_len_dtype),
            )
            request_output_len_tensor = pb_utils.Tensor(
                "REQUEST_OUTPUT_LEN", request_output_len
            )
            bad_words_ids_tensor = pb_utils.Tensor("BAD_WORDS_IDS", bad_words)
            stop_words_ids_tensor = pb_utils.Tensor("STOP_WORDS_IDS", stop_words)

            # Create InferenceResponse. You can set an error here in case
            # there was a problem with handling this inference request.
            # Below is an example of how you can set errors in inference
            # response:
            #
            # pb_utils.InferenceResponse(
            #    output_tensors=..., TritonError("An error occurred"))
            inference_response = pb_utils.InferenceResponse(
                output_tensors=[
                    input_id_tensor,
                    bad_words_ids_tensor,
                    stop_words_ids_tensor,
                    request_input_len_tensor,
                    request_output_len_tensor,
                ]
            )
            responses.append(inference_response)

        # You should return a list of pb_utils.InferenceResponse. Length
        # of this list must match the length of `requests` list.
        return responses

    def finalize(self):
        """`finalize` is called only once when the model is being unloaded.
        Implementing `finalize` function is optional. This function allows
        the model to perform any necessary clean ups before exit.
        """
        print("Cleaning up...")

    def _create_request(self, query):
        """
        query : batch string (2D numpy array)
        """
        start_ids = [
            np.array(
                self.tokenizer.encode(
                    s[0].decode(), add_special_tokens=self.add_special_tokens
                )
            ).astype(int)
            for s in query
        ]
        start_lengths = np.array([[len(ids)] for ids in start_ids]).astype(int)

        max_len = 0
        for seq in start_ids:
            max_len = max(max_len, seq.shape[0])
        start_ids = np.stack(
            [
                np.pad(
                    seq,
                    (0, max_len - seq.shape[0]),
                    "constant",
                    constant_values=(0, self.pad_id),
                )
                for seq in start_ids
            ]
        )

        return start_ids, start_lengths

    def _to_word_list_format(self, word_dict: List[List[str]]):
        """
        format of word_dict
            len(word_dict) should be same to batch_size
            word_dict[i] means the words for batch i
            len(word_dict[i]) must be 1, which means it only contains 1 string
            This string can contains several sentences and split by ",".
            For example, if word_dict[2] = " I am happy, I am sad", then this function will return
            the ids for two short sentences " I am happy" and " I am sad".
        """
        assert self.tokenizer is not None, "need to set tokenizer"

        flat_ids = []
        offsets = []
        for word_dict_item in word_dict:
            item_flat_ids = []
            item_offsets = []

            if isinstance(word_dict_item[0], bytes):
                word_dict_item = [word_dict_item[0].decode()]

            words = list(csv.reader(word_dict_item))[0]
            for word in words:
                ids = self.tokenizer.encode(word)

                if len(ids) == 0:
                    continue

                item_flat_ids += ids
                item_offsets.append(len(ids))

            flat_ids.append(np.array(item_flat_ids))
            offsets.append(np.cumsum(np.array(item_offsets)))

        pad_to = max(1, max(len(ids) for ids in flat_ids))

        for i, (ids, offs) in enumerate(zip(flat_ids, offsets)):
            flat_ids[i] = np.pad(ids, (0, pad_to - len(ids)), constant_values=0)
            offsets[i] = np.pad(offs, (0, pad_to - len(offs)), constant_values=-1)

        return np.array([flat_ids, offsets], dtype="int32").transpose((1, 0, 2))



from enum import Enum
from pathlib import Path
from typing import Optional

import numpy as np
import tritonclient
import tritonclient.grpc.aio as grpcclient
from pydantic import BaseModel, ConfigDict, PrivateAttr


class ModelInput:
    def __init__(
        self,
        prompt: str,
        request_id: int,
        max_tokens: int = 50,
        beam_width: int = 1,
        bad_words_list: Optional[list] = None,
        stop_words_list: Optional[list] = None,
        repetition_penalty: float = 1.0,
        ignore_eos: bool = False,
        stream: bool = True,
        eos_token_id: int = None,  # type: ignore
    ) -> None:
        self.stream = stream
        self.request_id = request_id
        self._prompt = prompt
        self._max_tokens = max_tokens
        self._beam_width = beam_width
        self._bad_words_list = [""] if bad_words_list is None else bad_words_list
        self._stop_words_list = [""] if stop_words_list is None else stop_words_list
        self._repetition_penalty = repetition_penalty
        self._eos_token_id = eos_token_id
        self._ignore_eos = ignore_eos

    def _prepare_grpc_tensor(
        self, name: str, input_data: np.ndarray
    ) -> grpcclient.InferInput:
        tensor = grpcclient.InferInput(
            name,
            input_data.shape,
            tritonclient.utils.np_to_triton_dtype(input_data.dtype),
        )
        tensor.set_data_from_numpy(input_data)
        return tensor

    def to_tensors(self):
        if self._eos_token_id is None and self._ignore_eos:
            raise ValueError("eos_token_id is required when ignore_eos is True")

        prompt_data = np.array([[self._prompt]], dtype=object)
        output_len_data = np.ones_like(prompt_data, dtype=np.uint32) * self._max_tokens
        bad_words_data = np.array([self._bad_words_list], dtype=object)
        stop_words_data = np.array([self._stop_words_list], dtype=object)
        stream_data = np.array([[self.stream]], dtype=bool)
        beam_width_data = np.array([[self._beam_width]], dtype=np.uint32)
        repetition_penalty_data = np.array(
            [[self._repetition_penalty]], dtype=np.float32
        )

        inputs = [
            self._prepare_grpc_tensor("text_input", prompt_data),
            self._prepare_grpc_tensor("max_tokens", output_len_data),
            self._prepare_grpc_tensor("bad_words", bad_words_data),
            self._prepare_grpc_tensor("stop_words", stop_words_data),
            self._prepare_grpc_tensor("stream", stream_data),
            self._prepare_grpc_tensor("beam_width", beam_width_data),
            self._prepare_grpc_tensor("repetition_penalty", repetition_penalty_data),
        ]

        if not self._ignore_eos:
            end_id_data = np.array([[self._eos_token_id]], dtype=np.uint32)
            inputs.append(self._prepare_grpc_tensor("end_id", end_id_data))

        return inputs


class Quant(Enum):
    NO_QUANT = "no_quant"
    WEIGHTS_ONLY = "weights_only"
    WEIGHTS_KV_INT8 = "weights_kv_int8"
    SMOOTH_QUANT = "smooth_quant"


class EngineType(Enum):
    LLAMA = "llama"
    MISTRAL = "mistral"


class ArgsConfig(BaseModel):
    max_input_len: Optional[int] = None
    max_output_len: Optional[int] = None
    max_batch_size: Optional[int] = None
    tp_size: Optional[int] = None
    pp_size: Optional[int] = None
    world_size: Optional[int] = None
    gather_all_token_logits: Optional[bool] = None
    multi_block_mode: Optional[bool] = None
    remove_input_padding: Optional[bool] = None
    use_gpt_attention_plugin: Optional[str] = None
    paged_kv_cache: Optional[bool] = None
    use_inflight_batching: Optional[bool] = None
    enable_context_fmha: Optional[bool] = None
    use_gemm_plugin: Optional[str] = None
    use_weight_only: Optional[bool] = None
    output_dir: Optional[str] = None
    model_dir: Optional[str] = None
    ft_model_dir: Optional[str] = None
    dtype: Optional[str] = None
    int8_kv_cache: Optional[bool] = None
    use_smooth_quant: Optional[bool] = None
    per_token: Optional[bool] = None
    per_channel: Optional[bool] = None
    parallel_build: Optional[bool] = None

    # to disable warning because `model_dir` starts with `model_` prefix
    model_config = ConfigDict(protected_namespaces=())  # type: ignore

    def as_command_arguments(self) -> list:
        non_bool_args = [
            element
            for arg, value in self.dict().items()
            for element in [f"--{arg}", str(value)]
            if value is not None and not isinstance(value, bool)
        ]
        bool_args = [
            f"--{arg}"
            for arg, value in self.dict().items()
            if isinstance(value, bool) and value
        ]
        return non_bool_args + bool_args


class CalibrationConfig(BaseModel):
    kv_cache: Optional[bool] = None  # either to calibrate kv cache
    sq_alpha: Optional[float] = None

    def cache_path(self) -> Path:
        if self.kv_cache is not None:
            return Path("kv_cache")
        else:
            return Path(f"sq_{self.sq_alpha}")


class EngineBuildArgs(BaseModel, use_enum_values=True):
    repo: Optional[str] = None
    args: Optional[ArgsConfig] = None
    quant: Optional[Quant] = None
    calibration: Optional[CalibrationConfig] = None
    engine_type: Optional[EngineType] = None


class TrussBuildConfig(BaseModel):
    """
    This is a spec for what the config.yaml looks like to take advantage of TRT-LLM + TRT-LLM builds. We structure the
    configuration with the below top-level keys.

    Example (for building an engine)
    ```
    build:
        model_server: TRT_LLM
        arguments:
            tokenizer_repository: "mistralai/mistral-v2-instruct"
            arguments:
                max_input_len: 1024
                max_output_len: 1024
                max_batch_size: 64
            quant: "weights_kv_int8"
            tensor_parallel_count: 2
            pipeline_parallel_count: 1
    ```

    Example (for using an existing engine)
    ```
    build:
        model_server: TRT_LLM
        arguments:
            engine_repository: "baseten/mistral-v2-32k"
            tensor_parallel_count: 2
            pipeline_parallel_count: 1
    ```

    """

    tokenizer_repository: str
    quant: Quant = Quant.NO_QUANT
    pipeline_parallel_count: int = 1
    tensor_parallel_count: int = 1
    arguments: Optional[ArgsConfig] = None
    engine_repository: Optional[str] = None
    calibration: Optional[CalibrationConfig] = None
    engine_type: Optional[EngineType] = None
    _engine_build_args: Optional[EngineBuildArgs] = PrivateAttr(default=None)

    @property
    def engine_build_args(self) -> EngineBuildArgs:
        if self._engine_build_args is None:
            repo = self.tokenizer_repository
            quant = self.quant
            calibration = self.calibration
            engine_type = self.engine_type
            args = self.arguments or ArgsConfig()
            args.tp_size = self.tensor_parallel_count
            args.pp_size = self.pipeline_parallel_count
            self._engine_build_args = EngineBuildArgs(
                repo=repo,
                quant=quant,
                calibration=calibration,
                engine_type=engine_type,
                args=args,
            )
        return self._engine_build_args

    @property
    def requires_build(self):
        return self.engine_repository is None



import os
from itertools import count

import build_engine_utils
from constants import (
    GRPC_SERVICE_PORT,
    HF_AUTH_KEY_CONSTANT,
    HTTP_SERVICE_PORT,
    TOKENIZER_KEY_CONSTANT,
)
from schema import ModelInput, TrussBuildConfig
from transformers import AutoTokenizer
from triton_client import TritonClient, TritonServer


class Model:
    def __init__(self, data_dir, config, secrets):
        self._data_dir = data_dir
        self._config = config
        self._secrets = secrets
        self._request_id_counter = count(start=1)
        self.triton_client = None
        self.triton_server = None
        self.tokenizer = None
        self.uses_openai_api = None

    def load(self):
        build_config = TrussBuildConfig(**self._config["build"]["arguments"])
        self.uses_openai_api = "openai-compatible" in self._config.get(
            "model_metadata", {}
        ).get("tags", [])
        hf_access_token = None
        if "hf_access_token" in self._secrets._base_secrets.keys():
            hf_access_token = self._secrets["hf_access_token"]

        # TODO(Abu): Move to pre-runtime
        if build_config.requires_build:
            build_engine_utils.build_engine_from_config_args(
                engine_build_args=build_config.engine_build_args,
                dst=self._data_dir,
            )

        self.triton_server = TritonServer(
            grpc_port=GRPC_SERVICE_PORT,
            http_port=HTTP_SERVICE_PORT,
        )

        self.triton_server.create_model_repository(
            truss_data_dir=self._data_dir,
            engine_repository_path=build_config.engine_repository
            if not build_config.requires_build
            else None,
            huggingface_auth_token=hf_access_token,
        )

        env = {}
        if hf_access_token:
            env[HF_AUTH_KEY_CONSTANT] = hf_access_token
        env[TOKENIZER_KEY_CONSTANT] = build_config.tokenizer_repository

        self.triton_server.start(
            tensor_parallelism=build_config.tensor_parallel_count,
            env=env,
        )

        self.triton_client = TritonClient(
            grpc_service_port=GRPC_SERVICE_PORT,
        )

        self.tokenizer = AutoTokenizer.from_pretrained(
            build_config.tokenizer_repository, token=hf_access_token
        )
        self.eos_token_id = self.tokenizer.eos_token_id

    async def predict(self, model_input):
        model_input["request_id"] = str(os.getpid()) + str(
            next(self._request_id_counter)
        )
        model_input["eos_token_id"] = self.eos_token_id

        self.triton_client.start_grpc_stream()

        model_input = ModelInput(**model_input)

        result_iterator = self.triton_client.infer(model_input)

        async def generate():
            async for result in result_iterator:
                yield result

        if model_input.stream:
            return generate()
        else:
            if self.uses_openai_api:
                return "".join(generate())
            else:
                return {"text": "".join(generate())}



# This file contains shared code to be used in other tests
# TODO(pankaj): Using a tests file for shared code is not ideal, we should
# move it to a regular file. This is a short term hack.

import shutil
import subprocess
import time
from contextlib import contextmanager

from truss.build import kill_all
from truss.constants import TRUSS
from truss.docker import get_containers

DISK_SPACE_LOW_PERCENTAGE = 20


@contextmanager
def ensure_kill_all():
    try:
        yield
    finally:
        kill_all_with_retries()
        ensure_free_disk_space()


def kill_all_with_retries(num_retries: int = 10):
    kill_all()
    attempts = 0
    while attempts < num_retries:
        containers = get_containers({TRUSS: True})
        if len(containers) == 0:
            return
        attempts += 1
        time.sleep(1)


def ensure_free_disk_space():
    """Check if disk space is low."""
    if is_disk_space_low():
        clear_disk_space()


def is_disk_space_low() -> bool:
    disk_usage = shutil.disk_usage("/")
    disk_free_percent = 100 * float(disk_usage.free) / disk_usage.total
    return disk_free_percent <= DISK_SPACE_LOW_PERCENTAGE


def clear_disk_space():
    docker_ps_output = subprocess.check_output(
        ["docker", "ps", "-a", "-f", "status=exited", "-q"]
    ).decode("utf-8")
    docker_containers = docker_ps_output.split("\n")[:-1]
    subprocess.run(["docker", "rm", *docker_containers])
    subprocess.run(["docker", "system", "prune", "-a", "-f"])



import datetime
import json
import os
import subprocess
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Type

import boto3
from botocore import UNSIGNED
from botocore.client import Config
from botocore.exceptions import ClientError, NoCredentialsError
from google.cloud import storage
from huggingface_hub import hf_hub_download

B10CP_PATH_TRUSS_ENV_VAR_NAME = "B10CP_PATH_TRUSS"

GCS_CREDENTIALS = "/app/data/service_account.json"
S3_CREDENTIALS = "/app/data/s3_credentials.json"


def _b10cp_path() -> Optional[str]:
    return os.environ.get(B10CP_PATH_TRUSS_ENV_VAR_NAME)


def _download_from_url_using_b10cp(
    b10cp_path: str,
    url: str,
    download_to: Path,
):
    return subprocess.Popen(
        [
            b10cp_path,
            "-source",
            url,  # Add quotes to work with any special characters.
            "-target",
            str(download_to),
        ]
    )


@dataclass
class AWSCredentials:
    access_key_id: str
    secret_access_key: str
    region: str
    session_token: Optional[str]


def parse_s3_credentials_file(key_file_path: str) -> AWSCredentials:
    # open the json file
    with open(key_file_path, "r") as f:
        data = json.load(f)

    # validate the data
    if (
        "aws_access_key_id" not in data
        or "aws_secret_access_key" not in data
        or "aws_region" not in data
    ):
        raise ValueError("Invalid AWS credentials file")

    # create an AWS Service Account object
    aws_sa = AWSCredentials(
        access_key_id=data["aws_access_key_id"],
        secret_access_key=data["aws_secret_access_key"],
        region=data["aws_region"],
        session_token=data.get("aws_session_token", None),
    )

    return aws_sa


def split_path(path, prefix="gs://"):
    # Remove the 'gs://' prefix
    path = path.replace(prefix, "")

    # Split on the first slash
    parts = path.split("/", 1)

    bucket_name = parts[0]
    path = parts[1] if len(parts) > 1 else ""

    return bucket_name, path


class RepositoryFile(ABC):
    def __init__(self, repo_name, file_name, revision_name):
        self.repo_name = repo_name
        self.file_name = file_name
        self.revision_name = revision_name

    @staticmethod
    def from_file(
        new_repo_name: str, new_file_name: str, new_revision_name: str
    ) -> "RepositoryFile":
        repository_class: Type["RepositoryFile"]
        if new_repo_name.startswith("gs://"):
            repository_class = GCSFile
        elif new_repo_name.startswith("s3://"):
            repository_class = S3File
        else:
            repository_class = HuggingFaceFile
        return repository_class(new_repo_name, new_file_name, new_revision_name)

    @abstractmethod
    def download_to_cache(self):
        pass


class HuggingFaceFile(RepositoryFile):
    def download_to_cache(self):
        secret_path = Path("/etc/secrets/hf-access-token")
        secret = secret_path.read_text().strip() if secret_path.exists() else None
        try:
            hf_hub_download(
                self.repo_name,
                self.file_name,
                revision=self.revision_name,
                token=secret,
            )
        except FileNotFoundError:
            raise RuntimeError(
                "Hugging Face repository not found (and no valid secret found for possibly private repository)."
            )


class GCSFile(RepositoryFile):
    def download_to_cache(self):
        # Create GCS Client
        bucket_name, _ = split_path(repo_name, prefix="gs://")

        is_private = os.path.exists(GCS_CREDENTIALS)
        print(is_private)
        if is_private:
            print("loading...")
            client = storage.Client.from_service_account_json(GCS_CREDENTIALS)
        else:
            client = storage.Client.create_anonymous_client()

        bucket = client.bucket(bucket_name)

        # Cache file
        cache_dir = Path(f"/app/model_cache/{bucket_name}")
        cache_dir.mkdir(parents=True, exist_ok=True)

        dst_file = cache_dir / self.file_name
        if not dst_file.parent.exists():
            dst_file.parent.mkdir(parents=True)

        blob = bucket.blob(self.file_name)

        if not blob.exists(client):
            raise RuntimeError(f"File not found on GCS bucket: {blob.name}")

        if is_private:
            url = blob.generate_signed_url(
                version="v4",
                expiration=datetime.timedelta(minutes=15),
                method="GET",
            )
        else:
            base_url = "https://storage.googleapis.com"
            url = f"{base_url}/{bucket_name}/{blob.name}"

        download_file_using_b10cp(url, dst_file, self.file_name)


class S3File(RepositoryFile):
    def download_to_cache(self):
        # Create S3 Client
        bucket_name, _ = split_path(repo_name, prefix="s3://")

        if os.path.exists(S3_CREDENTIALS):
            s3_credentials = parse_s3_credentials_file(S3_CREDENTIALS)
            client = boto3.client(
                "s3",
                aws_access_key_id=s3_credentials.access_key_id,
                aws_secret_access_key=s3_credentials.secret_access_key,
                region_name=s3_credentials.region,
                aws_session_token=s3_credentials.session_token,
                config=Config(signature_version="s3v4"),
            )
        else:
            client = boto3.client("s3", config=Config(signature_version=UNSIGNED))

        # Cache file
        cache_dir = Path(f"/app/model_cache/{bucket_name}")
        cache_dir.mkdir(parents=True, exist_ok=True)

        dst_file = cache_dir / self.file_name
        if not dst_file.parent.exists():
            dst_file.parent.mkdir(parents=True)

        try:
            url = client.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket_name, "Key": file_name},
                ExpiresIn=3600,
            )
        except NoCredentialsError as nce:
            raise RuntimeError(
                f"No AWS credentials found\nOriginal exception: {str(nce)}"
            )
        except ClientError as ce:
            raise RuntimeError(
                f"Client error when accessing the S3 bucket (check your credentials): {str(ce)}"
            )
        except Exception as exc:
            raise RuntimeError(
                f"File not found on S3 bucket: {file_name}\nOriginal exception: {str(exc)}"
            )

        download_file_using_b10cp(url, dst_file, self.file_name)


def download_file_using_b10cp(url, dst_file, file_name):
    try:
        proc = _download_from_url_using_b10cp(_b10cp_path(), url, dst_file)
        proc.wait()

    except FileNotFoundError as file_error:
        raise RuntimeError(f"Failure due to file ({file_name}) not found: {file_error}")


def download_file(repo_name, file_name, revision_name=None):
    file = RepositoryFile.from_file(repo_name, file_name, revision_name)
    file.download_to_cache()


if __name__ == "__main__":
    file_path = Path.home() / ".cache/huggingface/hub/version.txt"
    file_path.parent.mkdir(parents=True, exist_ok=True)

    if not file_path.is_file():
        file_path.write_text("1")

    file_name = sys.argv[1]
    repo_name = sys.argv[2]
    revision_name = sys.argv[3] if len(sys.argv) >= 4 else None

    download_file(repo_name, file_name, revision_name)



from pathlib import Path

from schema import EngineBuildArgs


def build_engine_from_config_args(
    engine_build_args: EngineBuildArgs,
    dst: Path,
):
    import os
    import shutil
    import sys

    # NOTE: These are provided by the underlying base image
    # TODO(Abu): Remove this when we have a better way of handling this
    sys.path.append("/app/baseten")
    from build_engine import Engine, build_engine
    from trtllm_utils import docker_tag_aware_file_cache

    engine = Engine(**engine_build_args.model_dump())

    with docker_tag_aware_file_cache("/root/.cache/trtllm"):
        built_engine = build_engine(engine, download_remote=True)

        if not os.path.exists(dst):
            os.makedirs(dst)

        for filename in os.listdir(str(built_engine)):
            source_file = os.path.join(str(built_engine), filename)
            destination_file = os.path.join(dst, filename)
            if not os.path.exists(destination_file):
                shutil.copy(source_file, destination_file)

        return dst



import importlib
import inspect
import sys
from pathlib import Path
from typing import Dict

import triton_python_backend_utils as pb_utils
from shared.secrets_resolver import SecretsResolver
from transform import transform_pydantic_to_triton, transform_triton_to_pydantic
from utils.errors import MissingInputClassError, MissingOutputClassError
from utils.signature import signature_accepts_keyword_arg


class TritonModelWrapper:
    def __init__(self, config: Dict):
        self._config = config
        self._model = None
        self._input_type = None
        self._output_type = None

    def instantiate(self) -> None:
        """
        Instantiates the model class defined in the config.
        """
        model_directory_prefix = Path("/app/model/1/truss")
        data_dir = model_directory_prefix / "data"

        if "bundled_packages_dir" in self._config:
            bundled_packages_path = model_directory_prefix / "bundled_packages"
            if bundled_packages_path.exists():
                sys.path.append(str(bundled_packages_path))
        model_module_name = str(
            Path(self._config["model_class_filename"]).with_suffix("")
        )
        module = importlib.import_module(
            f"truss.{self._config['model_module_dir']}.{model_module_name}"
        )
        model_class = getattr(module, self._config["model_class_name"])

        input_type_name = self._config.get("input_type_name", "Input")
        output_type_name = self._config.get("output_type_name", "Output")
        self._input_type = getattr(module, input_type_name, None)
        self._output_type = getattr(module, output_type_name, None)

        if not self._input_type:
            raise MissingInputClassError("Input type not defined in user model")
        if not self._output_type:
            raise MissingOutputClassError("Output type not defined in user model")

        model_class_signature = inspect.signature(model_class)
        model_init_params = {}
        if signature_accepts_keyword_arg(model_class_signature, "config"):
            model_init_params["config"] = self._config
        if signature_accepts_keyword_arg(model_class_signature, "data_dir"):
            model_init_params["data_dir"] = data_dir  # type: ignore
        if signature_accepts_keyword_arg(model_class_signature, "secrets"):
            model_init_params["secrets"] = SecretsResolver.get_secrets(self._config)

        self._model = model_class(**model_init_params)

    def load(self) -> None:
        """
        Invokes the underlying model's load method.
        """
        if self._model is None:
            raise ValueError("Model not instantiated")

        self._model.load()

    def predict(self, requests: list) -> list:
        """
        Converts Triton requests to Pydantic objects, invokes the underlying model's predict method, and converts the
        outputs to Triton objects.
        """
        requests = transform_triton_to_pydantic(requests, self._input_type)
        try:
            outputs = self._model.predict(requests)  # type: ignore
        except Exception as e:
            raise pb_utils.TritonModelException(e)

        return transform_pydantic_to_triton(outputs)



import importlib
import sys
import time
from pathlib import Path
from typing import Any

import pytest
import yaml


@pytest.fixture
def app_path(truss_container_fs: Path, helpers: Any):
    truss_container_app_path = truss_container_fs / "app"
    model_file_content = """
class Model:
    def __init__(self):
        self.load_count = 0
    def load(self):
        self.load_count += 1
        if self.load_count <= 2:
            raise RuntimeError('Simulated error')
    def predict(self, request):
        return request
    """
    with helpers.file_content(
        truss_container_app_path / "model" / "model.py",
        model_file_content,
    ), helpers.sys_path(truss_container_app_path):
        yield truss_container_app_path


# TODO: Make this test work
@pytest.mark.skip(
    reason="Succeeds when tests in this file are run alone, but fails with the whole suit"
)
def test_model_wrapper_load_error_once(app_path):
    if "model_wrapper" in sys.modules:
        model_wrapper_module = sys.modules["model_wrapper"]
        importlib.reload(model_wrapper_module)
    else:
        model_wrapper_module = importlib.import_module("model_wrapper")
    model_wraper_class = getattr(model_wrapper_module, "ModelWrapper")
    config = yaml.safe_load((app_path / "config.yaml").read_text())
    model_wrapper = model_wraper_class(config)
    model_wrapper.load()
    # Allow load thread to execute
    time.sleep(1)
    output = model_wrapper.predict({})
    assert output == {}
    assert model_wrapper._model.load_count == 3


# TODO: Make this test work
@pytest.mark.skip(
    reason="Succeeds when tests in this file are run alone, but fails with the whole suit"
)
def test_model_wrapper_load_error_more_than_allowed(app_path, helpers):
    with helpers.env_var("NUM_LOAD_RETRIES_TRUSS", "0"):
        if "model_wrapper" in sys.modules:
            model_wrapper_module = sys.modules["model_wrapper"]
            importlib.reload(model_wrapper_module)
        else:
            model_wrapper_module = importlib.import_module("model_wrapper")
        model_wraper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())
        model_wrapper = model_wraper_class(config)
        model_wrapper.load()
        # Allow load thread to execute
        time.sleep(1)
        assert model_wrapper.load_failed()


@pytest.mark.integration
async def test_model_wrapper_streaming_timeout(app_path):
    if "model_wrapper" in sys.modules:
        model_wrapper_module = sys.modules["model_wrapper"]
        importlib.reload(model_wrapper_module)
    else:
        model_wrapper_module = importlib.import_module("model_wrapper")
    model_wraper_class = getattr(model_wrapper_module, "ModelWrapper")

    # Create an instance of ModelWrapper with streaming_read_timeout set to 5 seconds
    config = yaml.safe_load((app_path / "config.yaml").read_text())
    config["runtime"]["streaming_read_timeout"] = 5
    model_wrapper = model_wraper_class(config)
    model_wrapper.load()
    assert model_wrapper._config.get("runtime").get("streaming_read_timeout") == 5



from typing import Optional


class Error(Exception):
    """Base Baseten Error"""

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class ApiError(Error):
    """Errors in calling the Baseten API."""

    def __init__(self, message: str, graphql_error_code: Optional[str] = None):
        super().__init__(message)
        self.graphql_error_code = graphql_error_code


class AuthorizationError(Error):
    """Raised in places where the user needs to be logged in and is not."""

    pass



from truss.patch.dir_signature import directory_content_signature


def test_directory_content_signature(tmp_path):
    root = tmp_path / "root"
    root.mkdir()
    (root / "file1").touch()
    (root / "file2").touch()
    subdir = root / "dir"
    subdir.mkdir()
    (subdir / "file3").touch()

    content_sign = directory_content_signature(root)

    assert content_sign.keys() == {
        "dir",
        "dir/file3",
        "file1",
        "file2",
    }


def test_directory_content_signature_no_data_dir(tmp_path):
    root = tmp_path / "root"
    root.mkdir()
    (root / "file1").touch()
    (root / "file2").touch()
    data_dir = root / "data"
    data_dir.mkdir()
    (data_dir / "file3").touch()

    content_sign = directory_content_signature(root=root, ignore_patterns=["data/*"])

    assert content_sign.keys() == {
        "data",
        "file1",
        "file2",
    }
